import os
import requests


class APIRepo:
    def __init__(self, url=None):
        default_url = "https://api.eotdl.com/"
        # default_url = "http://localhost:8000/"
        self.url = url if url else os.getenv("EOTDL_API_URL", default_url)

    def format_response(self, response):
        if response.status_code == 200:
            return response.json(), None
        return None, response.json()["detail"]

    def generate_headers(self, data):
        if "api_key" in data:
            return {"X-API-Key": data["api_key"]}
        if "id_token" in data:
            return {"Authorization": "Bearer " + data["id_token"]}
        raise Exception("Invalid headers")

    def get_info(self):
        response = requests.get(self.url)
        return self.format_response(response)



import typer
from typing import Optional

from ..curation.stac.api import api_status, search_stac_columns, retrieve_stac_collections, retrieve_stac_collection, retrieve_stac_items, retrieve_stac_item, search_stac_items

app = typer.Typer(help="EOTDL STAC API")

@app.command()
def status():
    try:
        data = api_status()
        typer.echo(data)
    except Exception as e:
        typer.echo(e)
        raise typer.Abort()

@app.command()
def collections():
    try:
        data = retrieve_stac_collections()
        typer.echo(data)
    except Exception as e:
        typer.echo(e)
        raise typer.Abort()
    
@app.command()
def collection(collection_id: str):
    try:
        data = retrieve_stac_collection(collection_id)
        typer.echo(data)
    except Exception as e:
        typer.echo(e)
        raise typer.Abort()
    
@app.command()
def items(collection_id: str):
    try:
        data = retrieve_stac_items(collection_id)
        typer.echo(data)
    except Exception as e:
        typer.echo(e)

@app.command()
def item(collection_id: str, item_id: str):
    try:
        data = retrieve_stac_item(collection_id, item_id)
        typer.echo(data)
    except Exception as e:
        typer.echo(e)

@app.command()
def search(collection_id: str, query: Optional[str] = None):
    try:
        data = search_stac_items(collection_id, query)
        typer.echo(data)
    except Exception as e:
        typer.echo(e)


"""
Time utils
"""

from datetime import datetime, timedelta
from typing import Union, Optional
from dateutil import parser

import geopandas as gpd
import pandas as pd


def is_time_interval(time_interval: list) -> bool:
    """
    Check if is time interval and is valid
    """
    if not isinstance(time_interval, (list, tuple)) or len(time_interval) != 2:
        return False

    for value in time_interval:
        if not isinstance(value, str):
            return False
        if not is_valid_date(value):
            return False

    return True


def is_valid_date(date_str: str) -> bool:
    """
    Check if a date is valid
    """
    try:
        datetime.strptime(date_str, "%Y-%m-%d")
        return True
    except ValueError:
        return False


def get_first_last_dates(
    dataframe: Union[pd.DataFrame, gpd.GeoDataFrame],
    dates_column: Optional[str] = "datetime",
):
    """
    Get first and last dates from a dataframe
    """
    dataframe[dates_column] = dataframe[dates_column].apply(lambda x: sorted(x))
    dataframe["first_date"] = dataframe["dates_list"].apply(lambda x: x[0])
    dataframe["last_date"] = dataframe["dates_list"].apply(lambda x: x[-1])
    dataframe = dataframe.sort_values(by=["first_date", "last_date"])
    # Sort by sequence id
    dataframe = dataframe.sort_values(by=["location_id"])
    # Convert first_date and last_date to datetime, in format YYYY-MM-DD
    dataframe["first_date"] = pd.to_datetime(dataframe["first_date"], format="%Y-%m-%d")
    dataframe["last_date"] = pd.to_datetime(dataframe["last_date"], format="%Y-%m-%d")

    return dataframe


def create_time_slots(start_date: datetime, end_date: datetime, n_chunks: int):
    """
    Create time slots from start date to end date, with n_chunks
    """
    if isinstance(start_date, str):
        start_date = datetime.strptime(start_date, "%Y-%m-%d")
    if isinstance(end_date, str):
        end_date = datetime.strptime(end_date, "%Y-%m-%d")
    tdelta = (end_date - start_date) / n_chunks
    edges = [(start_date + i * tdelta).date().isoformat() for i in range(n_chunks)]
    slots = [(edges[i], edges[i + 1]) for i in range(len(edges) - 1)]

    return slots


def expand_time_interval(
    time_interval: Union[list, tuple], time_format: str = "%Y-%m-%dT%H:%M:%S.%fZ"
) -> list:
    """
    Expand time interval to get more data
    """
    start_date = time_interval[0]
    end_date = time_interval[1]

    if isinstance(start_date, str):
        start_date = datetime.datetime.strptime(start_date, time_format)
    if isinstance(end_date, str):
        end_date = datetime.datetime.strptime(end_date, time_format)

    # Add one day to start date and remove one day to end date
    new_start_date = start_date - datetime.timedelta(days=1)
    new_end_date = end_date + datetime.timedelta(days=1)

    # Convert to string
    new_start_date = new_start_date.strftime(format)
    new_end_date = new_end_date.strftime(format)

    return new_start_date, new_end_date


def prepare_time_interval(date):
    """
    Prepare time interval to request data
    """
    if isinstance(date, str):
        date = datetime.strptime(date, "%Y-%m-%d")
    elif isinstance(date, tuple):
        if is_time_interval(date):
            return date
        else:
            raise ValueError(
                "The time interval must be a range of two dates, with format YYYY-MM-DD or a datetime object"
            )
    elif not isinstance(date, datetime):
        raise ValueError(
            "The date must be a string with format YYYY-MM-DD or a datetime object"
        )

    date_day_before = date - timedelta(days=1)
    date_next_day = date + timedelta(days=1)

    date_day_before_str = date_day_before.strftime("%Y-%m-%d")
    date_next_day_str = date_next_day.strftime("%Y-%m-%d")

    return (date_day_before_str, date_next_day_str)


def get_day_between(
    from_date: Union[datetime, str], to_date: Union[datetime, str]
) -> str:
    """
    Get the day between two dates
    """
    if isinstance(from_date, str):
        from_date = format_time_acquired(from_date)
    if isinstance(to_date, str):
        to_date = format_time_acquired(to_date)

    date_between = from_date + (to_date - from_date) / 2
    date_between = date_between.strftime("%Y-%m-%d")

    return date_between


def format_time_acquired(dt: Union[str, datetime]) -> str:
    """
    Format the date time to the required format for STAC

    :param dt: date time to format
    """
    dt_str = parser.parse(dt).strftime("%Y-%m-%dT%H:%M:%S.%f")
    # convert to datetime object
    dt = datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f")

    return dt



"""
Download imagery
"""

from datetime import datetime, timedelta
from typing import Union, List, Optional

from .sentinelhub import (
    SHClient,
    evaluate_sentinel_parameters,
    imagery_from_tmp_to_dir,
    get_default_parameters,
    SHParameters,
    filter_times,
)
from .search import advanced_imagery_search


def download_sentinel_imagery(
    output: str,
    time_interval: Union[str, datetime, List[Union[str, datetime]]],
    bounding_box: List[Union[int, float]],
    collection_id: str,
    name: Optional[str] = None,
) -> None:
    """
    Download Sentinel imagery
    """
    evaluate_sentinel_parameters(time_interval, bounding_box, collection_id, output)
    parameters = get_default_parameters(collection_id)
    return advanced_imagery_download(
        output, time_interval, bounding_box, parameters, name
    )


def advanced_imagery_download(
    output: str,
    time_interval: Union[str, datetime, List[Union[str, datetime]]],
    bounding_box: List[Union[int, float]],
    parameters: SHParameters,
    name: Optional[str] = None,
) -> None:
    """
    Advanced imagery download
    """
    evaluate_sentinel_parameters(
        time_interval, bounding_box, output=output, parameters=parameters
    )
    client = SHClient(parameters.BASE_URL)

    results = advanced_imagery_search(time_interval, bounding_box, parameters)
    timestamps = results.get_timestamps()
    time_difference = timedelta(hours=1)
    filtered_timestamps = filter_times(timestamps, time_difference)

    requests_list = []
    for date in filtered_timestamps:
        timestamp_interval = (date - time_difference, date + time_difference)
        requests_list.append(
            client.request_data(timestamp_interval, bounding_box, parameters)
        )
    if len(requests_list) == 0:
        print(
            f"No images found for {parameters.DATA_COLLECTION.api_id} in the specified time: {time_interval}"
        )
        return
    elif len(requests_list) <= 2:
        bulk = False
    else:
        bulk = True

    client.download_data(requests_list)
    imagery_from_tmp_to_dir(
        output,
        bounding_box,
        client.tmp_dir,
        name=name,
        bulk=bulk,
        output_format=parameters.OUTPUT_FORMAT.value,
    )


def search_and_download_sentinel_imagery(
    output: str,
    time_interval: Union[str, datetime, List[Union[str, datetime]]],
    bounding_box: List[Union[int, float]],
    sensor: str,
) -> None:
    """
    Search and download Sentinel imagery
    """
    from warnings import warn

    warn(
        "The function `search_and_download_sentinel_imagery` has been deprecated and will be removed in future updates. Please use download_satellite_imagery instead."
    )
    download_sentinel_imagery(output, time_interval, bounding_box, sensor)



"""
Metadata utilities for STAC
"""

import json

from os.path import dirname, join, exists, splitext
from os import listdir, remove
from glob import glob
from typing import Optional


def get_item_metadata(raster_path: str) -> str:
    """
    Get the metadata JSON file of a given directory, associated to a raster file

    :param raster_path: path to the raster file
    """
    # Get the directory of the raster file
    raster_dir_path = dirname(raster_path)
    # Get the metadata JSON file
    # Check if there is a metadata.json file in the directory
    if "metadata.json" in listdir(raster_dir_path):
        metadata_json = join(raster_dir_path, "metadata.json")
    else:
        # If there is no metadata.json file in the directory, check if there is
        # a json file with the same name as the raster file
        base = splitext(raster_path)[0]
        metadata_json = base + ".json"
        if not exists(metadata_json):
            # If there is no metadata file in the directory, return None
            return None

    # Open the metadata file and return it
    with open(metadata_json, "r", encoding="utf-8") as f:
        metadata = json.load(f)

    return metadata


def remove_raster_metadata(
    folder: str, metadata_file: Optional[str] = "metadata.json"
) -> None:
    """
    Remove metadata.json file from a folder

    :param folder: folder path
    :param metadata_file: metadata file name
    """
    # Search for all the metadata files in the folder
    metadata_files = glob(join(folder, "**", metadata_file), recursive=True)
    # Remove all the metadata files
    for metadata in metadata_files:
        remove(metadata)



# Q1+ model wrapper
# only works with some models, extend as we include more models in EOTDL and improve MLM extension

import os 
from pathlib import Path
from tqdm import tqdm
import numpy as np

from ..models.retrieve import retrieve_model
from ..curation.stac import STACDataFrame
from ..repos import FilesAPIRepo, ModelsAPIRepo
from ..auth import with_auth

class ModelWrapper:
    def __init__(self, model_name, version=None, path=None, force=False, assets=True, verbose=True):
        self.model_name = model_name
        self.version = version
        self.path = path
        self.force = force
        self.assets = assets
        self.verbose = verbose
        self.ready = False
        self.setup()

    def setup(self):
        download_path, gdf = self.download()
        self.download_path = download_path
        self.gdf = gdf
        # get model name from stac metadata
        item = gdf[gdf['type'] == "Feature"]
        assert item.shape[0] == 1, "Only one item is supported in stac metadata, found " + str(item.shape[0])
        self.props = item.iloc[0].properties
        assert self.props["mlm:framework"] == "ONNX", "Only ONNX models are supported, found " + self.props["mlm:framework"]
        model_name = self.props["mlm:name"]
        self.model_path = download_path + '/assets/' + model_name
        self.ready = True

    def predict(self, x):
        if not self.ready:
            self.setup()
        ort_session = self.get_onnx_session(self.model_path)
        # preprocess input
        x = self.process_inputs(x)
        # execute model
        input_name = ort_session.get_inputs()[0].name
        ort_inputs = {input_name: x}
        ort_outs = ort_session.run(None, ort_inputs)
        output_nodes = ort_session.get_outputs()
        output_names = [node.name for node in output_nodes]
        # format and return outputs
        return self.return_outputs(ort_outs, output_names)

    @with_auth
    def download(self, user=None):
        # download the model
        model = retrieve_model(self.model_name)
        if model["quality"] == 0:
            raise Exception("Only Q1+ models are supported")
        if self.version is None:
            self.version = sorted(model["versions"], key=lambda v: v["version_id"])[-1][
                "version_id"
            ]
        else:
            assert self.version in [
                v["version_id"] for v in model["versions"]
            ], f"Version {self.version} not found"
        download_base_path = os.getenv(
            "EOTDL_DOWNLOAD_PATH", str(Path.home()) + "/.cache/eotdl/models"
        )
        if self.path is None:
            download_path = download_base_path + "/" + self.model_name + "/v" + str(self.version)
        else:
            download_path = self.path + "/" + self.model_name + "/v" + str(self.version)
        # check if model already exists
        if os.path.exists(download_path) and not self.force:
            os.makedirs(download_path, exist_ok=True)
            gdf = STACDataFrame.from_stac_file(download_path + f"/{self.model_name}/catalog.json")
            return download_path, gdf
        if self.verbose:
            print("Downloading STAC metadata...")
        repo = ModelsAPIRepo()
        gdf, error = repo.download_stac(
            model["id"],
            user,
        )
        if error:
            raise Exception(error)
        df = STACDataFrame(gdf)
        # df.geometry = df.geometry.apply(lambda x: Polygon() if x is None else x)
        df.to_stac(download_path)
        # download assets
        if self.assets:
            if self.verbose:
                print("Downloading assets...")
            repo = FilesAPIRepo()
            df = df.dropna(subset=["assets"])
            for row in tqdm(df.iterrows(), total=len(df)):
                for k, v in row[1]["assets"].items():
                    href = v["href"]
                    _, filename = href.split("/download/")
                    # will overwrite assets with same name :(
                    repo.download_file_url(
                        href, filename, f"{download_path}/assets", user
                    )
        else:
            print("To download assets, set assets=True.")
        if self.verbose:
            print("Done")
        return download_path, gdf

    def process_inputs(self, x):
        # pre-process and validate input
        input = self.props["mlm:input"]
        # input data type
        dtype = input["input"]["data_type"]
        x = x.astype(dtype)
        # input shape
        input_shape = input["input"]["shape"]
        ndims = len(input_shape)
        if ndims != x.ndim:
            if ndims == 4:
                x = np.expand_dims(x, axis=0).astype(np.float32)
            else:
                raise Exception("Input shape not valid", input_shape, x.ndim)
        for i, dim in enumerate(input_shape):
            if dim != -1:
                assert dim == x.shape[i], f"Input dimension not valid: The model expects {input_shape} but input has {x.shape} (-1 means any dimension)."
        # TODO: should apply normalization if defined in metadata
        return x
    
    def return_outputs(self, ort_outputs, output_names):
        if self.props["mlm:output"]["tasks"] == ["classification"]:
            return {
                "model": self.model_name,
                **{
                    output: ort_outputs[i].tolist() for i, output in enumerate(output_names)
                },
            }
        elif self.props["mlm:output"]["tasks"] == ["segmentation"]:
            outputs = {output: ort_outputs[i] for i, output in enumerate(output_names)}
            batch = outputs[output_names[0]]
            image = batch[0]
            return image
        else:
            raise Exception("Output task not supported:", self.props["mlm:output"]["tasks"])

    def get_onnx_session(self, model):        
        try:
            import onnxruntime as ort
            # gpu requires `pip install onnxruntime-gpu` but no extra imports
        except ImportError:
            raise ImportError("onnxruntime is not installed. Please install it with `pip install onnxruntime`")
        providers = ["CUDAExecutionProvider", "CPUExecutionProvider"]
        try:
            session = ort.InferenceSession(model, providers=providers)
        except Exception as e:
            raise RuntimeError(f"Error loading ONNX model: {str(e)}")
        return session



from pydantic import BaseModel, validator
from typing import List, Optional
from pathlib import Path
import os

class Metadata(BaseModel):
    authors: List[str]
    license: str
    source: str
    name: str
    description: str
    thumbnail: Optional[str] = ""

    # validate source is a URL
    @validator("source")
    def source_is_url(cls, v):
        if not v.startswith("http") and not v.startswith("https"):
            raise ValueError("source must be a URL")
        return v

    # validate thumbnail is a url
    @validator("thumbnail")
    def thumbnail_is_url(cls, v):
        if not v:
            return ""
        if not v.startswith("http") and not v.startswith("https"):
            raise ValueError("thumbnail must be a URL")
        return v


    def save_metadata(self, dst_path):
        os.makedirs(dst_path, exist_ok=True)
        with open(Path(dst_path) / "README.md", "w") as f:
            f.write("---\n")
            f.write(f"name: {self.name}\n")
            f.write(f"license: {self.license}\n") 
            f.write(f"source: {self.source}\n")
            f.write(f"thumbnail: {self.thumbnail}\n")
            f.write(f"authors:\n")
            for author in self.authors:
                f.write(f"  - {author}\n")
            f.write("---\n")
            f.write(self.description)
        return str(Path(dst_path) / "README.md")


import requests

from ..repos import APIRepo


class STACAPIRepo(APIRepo):
    def __init__(self, url=None):
        super().__init__(url)

    def status(self):
        response = requests.get(self.url + "stac")
        return self.format_response(response)
    
    def collections(self):
        response = requests.get(self.url + "stac/collections")
        return self.format_response(response)

    def collection(self, collection_id):
        response = requests.get(self.url + f"stac/collections/{collection_id}")
        return self.format_response(response)

    def items(self, collection_id):
        response = requests.get(self.url + f"stac/collections/{collection_id}/items")
        return self.format_response(response)

    def item(self, collection_id, item_id):
        response = requests.get(self.url + f"stac/collections/{collection_id}/items/{item_id}")
        return self.format_response(response)
    
    def search(self, collection_id, query):
        body = {"collection_id": collection_id}
        if query is not None:
            body["query"] = query
        response = requests.post(self.url + f"stac/search", json=body)
        return self.format_response(response)
    
    def search_columns(self, collection_id):
        response = requests.get(self.url + f"stac/search?collection={collection_id}")
        return self.format_response(response)


import pyarrow.parquet as pq
import stac_geoparquet
import json
from tqdm import tqdm
import pystac
from datetime import datetime

def create_stac_catalog(parquet_catalog_path, stac_catalog = None):
	# parse items and add to collection
	table = pq.read_table(parquet_catalog_path)
	items = []
	for item in tqdm(stac_geoparquet.arrow.stac_table_to_items(table), total=len(table)):
		item = pystac.Item.from_dict(item)
		# item.validate()
		# collection.add_item(item)
		if stac_catalog is not None:
			stac_catalog.add_item(item)
		else:
			items.append(item)
		# path = "data/stac/" + item["id"] + ".json"
		# os.makedirs(os.path.dirname(path), exist_ok=True)
		# with open(path, "w") as f:
		# 	json.dump(item, f)
		# # save item
		# os.makedirs(path, exist_ok=True)
		# _path = path + '/' + item.id + ".json"
		# os.makedirs(os.path.dirname(_path), exist_ok=True)
		# with open(_path, "w") as f:
		# 	json.dump(item.to_dict(), f)
	# save catalog
	if stac_catalog is not None:
		return stac_catalog
	return items


from .retrieve import retrieve_models, retrieve_model, retrieve_model_files
from .ingest import ingest_model, ingest_virtual_model, ingest_model_catalog
from .stage import stage_model, stage_model_file



"""
Data access module for eotdl package.
"""

from .download import (
    download_sentinel_imagery,
    search_and_download_sentinel_imagery,
    advanced_imagery_download,
)
from .search import search_sentinel_imagery, advanced_imagery_search
from .sentinelhub.parameters import (
    SUPPORTED_COLLECTION_IDS,
    DATA_COLLECTION_ID,
    get_default_parameters,
    OUTPUT_FORMAT,
)
from .sentinelhub.evalscripts import EvalScripts



"""
Evalscripts for Sentinel Hub requests
"""


class EvalScripts:
    """
    Class that defines the needed Sentinel Hub evalscripts
    """

    SENTINEL_1 = """
                //VERSION=3
                function setup() {
                    return {
                        input: [{
                            bands: ["VH", "VV"]
                        }],
                        output: {
                            id: "default",
                            bands: 2
                        }
                    };
                }

                function evaluatePixel(sample) {
                    return [sample.VH, sample.VV];
                }
                """

    SENTINEL_1_ENHANCED_VISUALIZATION = """
                //VERSION=3
                function setup() {
                return {
                    input: ["VV", "VH", "dataMask"],
                    output: { bands: 4 }
                }
                }

                function evaluatePixel(sample) {
                var water_threshold = 25; //lower means more water
                
                if (sample.VV / sample.VH > water_threshold) {
                    // watervis
                    return [sample.VV, 8 * sample.VV, 0.5 + 3 * sample.VV + 2000 * sample.VH, sample.dataMask];
                } else {
                    // landvis
                    return [3 * sample.VV, 1.1 * sample.VV + 8.75 * sample.VH, 1.75 * sample.VH, sample.dataMask];
                }
                }
                """

    SENTINEL_2_L1C = """
                //VERSION=3
                function setup() {
                    return {
                        input: [{
                            bands: ["B01",
                                    "B02",
                                    "B03",
                                    "B04",
                                    "B05",
                                    "B06",
                                    "B07",
                                    "B08",
                                    "B8A",
                                    "B09",
                                    "B10",
                                    "B11",
                                    "B12"],
                            units: "DN"
                        }],
                        output: {
                            bands: 13,
                            sampleType: "INT16"
                        }
                    };
                }

                function evaluatePixel(sample) {
                    return [sample.B01,
                            sample.B02,
                            sample.B03,
                            sample.B04,
                            sample.B05,
                            sample.B06,
                            sample.B07,
                            sample.B08,
                            sample.B8A,
                            sample.B09,
                            sample.B10,
                            sample.B11,
                            sample.B12];
                }
                """

    SENTINEL_2_L2A = """
                //VERSION=3
                function setup() {
                    return {
                        input: [{
                            bands: ["B01",
                                    "B02",
                                    "B03",
                                    "B04",
                                    "B05",
                                    "B06",
                                    "B07",
                                    "B08",
                                    "B8A",
                                    "B09",
                                    "B11",
                                    "B12"],
                            units: "DN"
                        }],
                        output: {
                            bands: 12,
                            sampleType: "INT16"
                        }
                    };
                }

                function evaluatePixel(sample) {
                    return [sample.B01,
                            sample.B02,
                            sample.B03,
                            sample.B04,
                            sample.B05,
                            sample.B06,
                            sample.B07,
                            sample.B08,
                            sample.B8A,
                            sample.B09,
                            sample.B11,
                            sample.B12];
                }
                """

    SENTINEL_2_L2A_TRUE_COLOR = """
        //VERSION=3
        let minVal = 0.0;
        let maxVal = 0.4;

        let viz = new HighlightCompressVisualizer(minVal, maxVal);

        function setup() {
        return {
            input: ["B04", "B03", "B02","dataMask"],
            output: { bands: 4 }
        };
        }

        function evaluatePixel(samples) {
            let val = [samples.B04, samples.B03, samples.B02,samples.dataMask];
            return viz.processList(val);
        }
        """

    DEM = """
        //VERSION=3

        function setup() {
            return {
                input: ["DEM"],
                output: { id: "default",
                        bands: 1,
                        sampleType: SampleType.FLOAT32
                },
            }
        }

        function evaluatePixel(sample) {
            return [sample.DEM]
        }
        """

    DEM_TOPOGRAPHIC = """
        //VERSION=3
        // To set custom max and min values, set
        // defaultVis to false and choose your max and
        // min values. The color map will then be scaled
        // to those max and min values
        const defaultVis = true
        const max = 9000
        const min = -9000

        function setup() {
            return {
                input: ["DEM", "dataMask"],
                output: [
                    { id: "default", bands: 4, sampleTYPE: 'AUTO' },
                    { id: "index", bands: 1, sampleType: 'FLOAT32' },
                    { id: "dataMask", bands: 1 }
                ]
            };
        }

        function updateMap(max, min) {
            const numIntervals = map.length
            const intervalLength = (max - min) / (numIntervals - 1);
            for (let i = 0; i < numIntervals; i++) {
                map[i][0] = max - intervalLength * i
            }
        }

        const map = [
            [8000, 0xffffff],
            [7000, 0xf2f2f2],
            [6000, 0xe5e5e5],
            [5500, 0x493811],
            [5000, 0x5e4c26],
            [4500, 0x726038],
            [4000, 0x87724c],
            [3500, 0x998760],
            [3000, 0xad9b75],
            [2500, 0xc1af89],
            [2000, 0xd6c49e],
            [1500, 0xead8af],
            [1000, 0xfcedbf],
            [900, 0xaadda0],
            [800, 0xa5d69b],
            [700, 0x96ce8e],
            [600, 0x84c17a],
            [500, 0x7aba70],
            [400, 0x72b266],
            [300, 0x5ea354],
            [200, 0x4c933f],
            [100, 0x3d873d],
            [75, 0x357c3a],
            [50, 0x2d722d],
            [25, 0x266821],
            [10, 0x1e5e14],
            [0.01, 0x165407]
        ];

        if (!defaultVis) updateMap(max, min);
        // add ocean color
        map.push([-10000, 0x0f0f8c])
        const visualizer = new ColorMapVisualizer(map);

        function evaluatePixel(sample) {
            let val = sample.DEM;
            let imgVals = visualizer.process(val)

            // Return the 4 inputs and define content for each one
            return {
                default: [...imgVals, sample.dataMask],
                index: [val],
                dataMask: [sample.dataMask]
            };
        }
        """

    HLS_FALSE_COLOR = """
        //VERSION=3

        function setup() {
        return {
            input: ["NIR_Narrow", "Red", "Green", "dataMask"],
            output: { bands: 3 }
        };
        }

        function evaluatePixel(sample) {
        return [2.5 * sample.NIR_Narrow, 2.5 * sample.Red, 2.5 * sample.Green, sample.dataMask];
        }
        """

    HLS_TRUE_COLOR = """
        //VERSION=3

        function setup() {
            return {
                input: ["Blue","Green","Red", "dataMask"],
                output: { bands: 4 }
            };
        }

        function evaluatePixel(sample) {
            return [2.5 * sample.Red, 2.5 * sample.Green, 2.5 * sample.Blue, sample.dataMask];
        }
        """

    HLS_NDVI = """
        //VERSION=3

        function setup() { 
            return {
                input: ["NIR_Narrow", "Red", "dataMask"],
                output: { bands: 4 }
            };
        }

        function evaluatePixel(sample) {
            var ndvi = (sample.NIR_Narrow - sample.Red) / (sample.NIR_Narrow + sample.Red)

            if (ndvi<-1.1) return [0,0,0, sample.dataMask];
            else if (ndvi<-0.2) return [0.75,0.75,0.75, sample.dataMask];
            else if (ndvi<-0.1) return [0.86,0.86,0.86, sample.dataMask];
            else if (ndvi<0) return [1,1,0.88, sample.dataMask];
            else if (ndvi<0.025) return [1,0.98,0.8, sample.dataMask];
            else if (ndvi<0.05) return [0.93,0.91,0.71, sample.dataMask];
            else if (ndvi<0.075) return [0.87,0.85,0.61, sample.dataMask];
            else if (ndvi<0.1) return [0.8,0.78,0.51, sample.dataMask];
            else if (ndvi<0.125) return [0.74,0.72,0.42, sample.dataMask];
            else if (ndvi<0.15) return [0.69,0.76,0.38, sample.dataMask];
            else if (ndvi<0.175) return [0.64,0.8,0.35, sample.dataMask];
            else if (ndvi<0.2) return [0.57,0.75,0.32, sample.dataMask];
            else if (ndvi<0.25) return [0.5,0.7,0.28, sample.dataMask];
            else if (ndvi<0.3) return [0.44,0.64,0.25, sample.dataMask];
            else if (ndvi<0.35) return [0.38,0.59,0.21, sample.dataMask];
            else if (ndvi<0.4) return [0.31,0.54,0.18, sample.dataMask];
            else if (ndvi<0.45) return [0.25,0.49,0.14, sample.dataMask];
            else if (ndvi<0.5) return [0.19,0.43,0.11, sample.dataMask];
            else if (ndvi<0.55) return [0.13,0.38,0.07, sample.dataMask];
            else if (ndvi<0.6) return [0.06,0.33,0.04, sample.dataMask];
            else return [0,0.27,0, sample.dataMask];
        }
        """

    LANDSAT_OT_L2_TRUE_COLOR = """
        //VERSION=3

        let minVal = 0.0;
        let maxVal = 0.4;

        let viz = new DefaultVisualizer(minVal, maxVal);

        function evaluatePixel(samples) {
            let val = [samples.B04, samples.B03, samples.B02, samples.dataMask];
            return viz.processList(val);
        }

        function setup() {
            return {
                input: [{
                bands: [ "B02", "B03", "B04", "dataMask" ]
                }],
                output: { bands: 4 }  }
            }
        """
    LANDSAT_OT_L2_NDVI = """
        //VERSION=3
        
        function setup() {
        return {
            input: ["B03", "B04", "B05", "dataMask"],
            output: [
            { id: "default", bands: 4 },
            { id: "index", bands: 1, sampleType: "FLOAT32" },
            { id: "eobrowserStats", bands: 2 },
            { id: "dataMask", bands: 1 },
            ],
        };
        }

        function evaluatePixel(samples) {
        let val = index(samples.B05, samples.B04);
        let imgVals = null;
        // The library for tiffs works well only if there is only one channel returned.
        // So we encode the "no data" as NaN here and ignore NaNs on frontend.
        const indexVal = samples.dataMask === 1 && val >= -1 && val <= 1 ? val : NaN;

        if (val < -1.1) imgVals = [0, 0, 0, samples.dataMask];
        else if (val < -0.2) imgVals = [0.75, 0.75, 0.75, samples.dataMask];
        else if (val < -0.1) imgVals = [0.86, 0.86, 0.86, samples.dataMask];
        else if (val < 0) imgVals = [1, 1, 0.88, samples.dataMask];
        else if (val < 0.025) imgVals = [1, 0.98, 0.8, samples.dataMask];
        else if (val < 0.05) imgVals = [0.93, 0.91, 0.71, samples.dataMask];
        else if (val < 0.075) imgVals = [0.87, 0.85, 0.61, samples.dataMask];
        else if (val < 0.1) imgVals = [0.8, 0.78, 0.51, samples.dataMask];
        else if (val < 0.125) imgVals = [0.74, 0.72, 0.42, samples.dataMask];
        else if (val < 0.15) imgVals = [0.69, 0.76, 0.38, samples.dataMask];
        else if (val < 0.175) imgVals = [0.64, 0.8, 0.35, samples.dataMask];
        else if (val < 0.2) imgVals = [0.57, 0.75, 0.32, samples.dataMask];
        else if (val < 0.25) imgVals = [0.5, 0.7, 0.28, samples.dataMask];
        else if (val < 0.3) imgVals = [0.44, 0.64, 0.25, samples.dataMask];
        else if (val < 0.35) imgVals = [0.38, 0.59, 0.21, samples.dataMask];
        else if (val < 0.4) imgVals = [0.31, 0.54, 0.18, samples.dataMask];
        else if (val < 0.45) imgVals = [0.25, 0.49, 0.14, samples.dataMask];
        else if (val < 0.5) imgVals = [0.19, 0.43, 0.11, samples.dataMask];
        else if (val < 0.55) imgVals = [0.13, 0.38, 0.07, samples.dataMask];
        else if (val < 0.6) imgVals = [0.06, 0.33, 0.04, samples.dataMask];
        else imgVals = [0, 0.27, 0, samples.dataMask];

        return {
            default: imgVals,
            index: [indexVal],
            eobrowserStats: [val, isCloud(samples) ? 1 : 0],
            dataMask: [samples.dataMask],
        };
        }

        function isCloud(samples) {
        const NGDR = index(samples.B03, samples.B04);
        const bRatio = (samples.B03 - 0.175) / (0.39 - 0.175);
        return bRatio > 1 || (bRatio > 0 && NGDR > 0);
        }
        """
